from collections import namedtuple

__version__ = "0.2.18"

version_info = namedtuple("version_info", ["major", "minor", "micro", "releaselevel"])

major_minor_micro = __version__.split("-")[0]

s = major_minor_micro.split(".")

version_info.major = int(s[0])
version_info.minor = int(s[1])
version_info.micro = int(s[0])

version_info.releaselevel = __version__[len(major_minor_micro) + 1 :]



import pandas as pd
import numpy as np

NoneType = type(None)

def is_int(*args):
    if not all([isinstance(arg, int) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, int)])
        raise ValueError(f'The object is expected to be a integer, not {types}')

def is_int_or_none(*args):
    if not all([isinstance(arg, (int, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, int)])
        raise ValueError(f'The object is expected to be a integer, not {types}')

def is_float(*args):
    if not all([isinstance(arg, float) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, float)])
        raise ValueError(f'The object is expected to be a float, not {types}')

def is_float_or_none(*args):
    if not all([isinstance(arg, (float, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, float)])
        raise ValueError(f'The object is expected to be a float, not {types}')

def is_boolean(*args):
    if not all([isinstance(arg, bool) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, bool)])
        raise ValueError(f'The object is expected to be a boolean, not {types}')

def is_boolean_or_none(*args):
    if not all([isinstance(arg, (bool, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, bool)])
        raise ValueError(f'The object is expected to be a boolean, not {types}')

def is_str(*args):
    if not all([isinstance(arg, str) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, str)])
        raise ValueError(f'The object is expected to be a string, not {types}')

def is_str_or_none(*args):
    if not all([isinstance(arg, (str, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, str)])
        raise ValueError(f'The object is expected to be a string, not {types}')

def is_list(*args):
    if not all([isinstance(arg, list) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, list)])
        raise ValueError(f'The object is expected to be a list, not {types}')

def is_list_or_none(*args):
    if not all([isinstance(arg, (list, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, list)])
        raise ValueError(f'The object is expected to be a list, not {types}')

def is_tuple(*args):
    if not all([isinstance(arg, tuple) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, tuple)])
        raise ValueError(f'The object is expected to be a tuple, not {types}')

def is_tuple_or_none(*args):
    if not all([isinstance(arg, (tuple, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, tuple)])
        raise ValueError(f'The object is expected to be a tuple, not {types}')

def is_array(*args):
    if not all([isinstance(arg, np.ndarray) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, np.ndarray)])
        raise ValueError(f'The object is expected to be a numpy ndarray, not {types}')
    
def is_array_or_none(*args):
    if not all([isinstance(arg, (np.ndarray, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, np.ndarray)])
        raise ValueError(f'The object is expected to be a numpy ndarray, not {types}')
    
def is_list_or_tuple(*args):
    if not all([isinstance(arg, (list, tuple)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, (list, tuple))])
        raise ValueError(f'The object is expected to be a list or a tuple, not {types}')
is_tuple_or_list = is_list_or_tuple

def is_list_or_tuple_or_none(*args):
    if not all([isinstance(arg, (list, tuple, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, (list, tuple))])
        raise ValueError(f'The object is expected to be a list or a tuple, not {types}')
is_tuple_or_list_or_none = is_list_or_tuple_or_none


def is_list_or_tuple_or_array(*args):
    if not all([isinstance(arg, (list, tuple, np.ndarray)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, (list, tuple, np.ndarray))])
        raise ValueError(f'The object is expected to be a list, tuple or numpy ndarray, not {types}')
is_list_or_array_or_tuple = is_list_or_tuple_or_array
is_tuple_or_list_or_array = is_list_or_tuple_or_array
is_tuple_or_array_or_list = is_list_or_tuple_or_array
is_array_or_tuple_or_list = is_list_or_tuple_or_array
is_array_or_list_or_tuple = is_list_or_tuple_or_array

def is_list_or_tuple_or_array_or_none(*args):
    if not all([isinstance(arg, (list, tuple, np.ndarray, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, (list, tuple, np.ndarray))])
        raise ValueError(f'The object is expected to be a list, tuple or numpy ndarray, not {types}')
is_list_or_array_or_tuple_or_none = is_list_or_tuple_or_array_or_none
is_tuple_or_array_or_list_or_none = is_list_or_tuple_or_array_or_none
is_array_or_tuple_or_list_or_none = is_list_or_tuple_or_array_or_none
is_array_or_list_or_tuple_or_none = is_list_or_tuple_or_array_or_none

def is_dict(*args):
    if not all([isinstance(arg, dict) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, dict)])
        raise ValueError(f'The object is expected to be a dictionary, not {types}')

def is_dict_or_none(*args):
    if not all([isinstance(arg, (dict, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, dict)])
        raise ValueError(f'The object is expected to be a dictionary, not {types}')

def is_series(*args):
    if not all([isinstance(arg, pd.Series) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.Series)])
        raise ValueError(f'The object is expected to be a pandas Series, not {types}')

def is_series_or_none(*args):
    if not all([isinstance(arg, (pd.Series, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.Series)])
        raise ValueError(f'The object is expected to be a pandas Series, not {types}')

def is_frame(*args):
    if not all([isinstance(arg, pd.DataFrame) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.DataFrame)])
        raise ValueError(f'The object is expected to be a pandas DataFrame, not {types}')

def is_frame_or_none(*args):
    if not all([isinstance(arg, (pd.DataFrame, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.DataFrame)])
        raise ValueError(f'The object is expected to be a pandas DataFrame, not {types}')

def is_series_or_frame(*args):
    if not all([isinstance(arg, (pd.Series, pd.DataFrame)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, (pd.Series, pd.DataFrame))])
        raise ValueError(f'The object is expected to be a pandas Series or DataFrame, not {types}')
is_frame_or_series = is_series_or_frame

def is_series_or_frame_or_none(*args):
    if not all([isinstance(arg, (pd.Series, pd.DataFrame, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, (pd.Series, pd.DataFrame))])
        raise ValueError(f'The object is expected to be a pandas Series or DataFrame, not {types}')
is_frame_or_series_or_none = is_series_or_frame_or_none

def is_index(*args):
    if not all([isinstance(arg, pd.Index) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.Index)])
        raise ValueError(f'The object is expected to be a pandas Index, not {types}')

def is_index_or_none(*args):
    if not all([isinstance(arg, (pd.Index, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.Index)])
        raise ValueError(f'The object is expected to be a pandas Index, not {types}')

def is_datetimeindex(*args):
    if not all([isinstance(arg, pd.DatetimeIndex) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.DatetimeIndex)])
        raise ValueError(f'The object is expected to be a pandas DatetimeIndex, not {types}')

def is_datetimeindex_or_none(*args):
    if not all([isinstance(arg, (pd.DatetimeIndex, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.DatetimeIndex)])
        raise ValueError(f'The object is expected to be a pandas DatetimeIndex, not {types}')

def is_multiindex(*args):
    if not all([isinstance(arg, pd.MultiIndex) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.MultiIndex)])
        raise ValueError(f'The object is expected to be a pandas MultiIndex, not {types}')

def is_multiindex_or_none(*args):
    if not all([isinstance(arg, (pd.MultiIndex, NoneType)) for arg in args]):
        types = ', '.join([type(arg).__name__ for arg in args if not isinstance(arg, pd.MultiIndex)])
        raise ValueError(f'The object is expected to be a pandas MultiIndex, not {types}')

def has_datetimeindex(*args):
    if not all([isinstance(arg.index, pd.DatetimeIndex) for arg in args]):
        types = ', '.join([type(arg.index).__name__ for arg in args if not isinstance(arg.index, pd.DatetimeIndex)])
        raise ValueError(f'The object is expected to have a pandas DateTimeIndex, not {types}')

def has_multiindex(*args):
    if not all([isinstance(arg.index, pd.MultiIndex) for arg in args]):
        types = ', '.join([type(arg.index).__name__ for arg in args if not isinstance(arg.index, pd.MultiIndex)])
        raise ValueError(f'The object is expected to have a MultiIndex, not {types}')


import numpy as np
import pandas as pd
import datetime


def reduce(_obj, *args, **kwargs):
    """
        Reduce the pandas object by grouping the index by value.
    
        Parameters
        ----------
        _obj: pandas Series or DataFrame

        Returns
        -------
        A pandas groupby object

        Example
        --------
        >>> index = pd.date_range('2022-01-01 12:00:00', periods=7, freq='15T')
        >>> data = [1., 2., 2., 3., 2., 2., 0.]]
        >>> series = pd.Series(index=index, data=data)
        >>> series
        2022-01-01 12:00:00    1.0
        2022-01-01 12:15:00    2.0
        2022-01-01 12:30:00    2.0
        2022-01-01 12:45:00    3.0
        2022-01-01 13:00:00    2.0
        2022-01-01 13:15:00    2.0
        2022-01-01 13:30:00    0.0
        dtype: float64
        
        Reduce the series
        >>> series._.reduce()
        2022-01-01 12:00:00    1.0
        2022-01-01 12:15:00    2.0
        2022-01-01 12:45:00    3.0
        2022-01-01 13:00:00    2.0
        2022-01-01 13:30:00    0.0
        dtype: float64

    """
    index = _obj.index[(_obj.values != _obj.shift().values)]
    data = _obj.groupby((_obj.values != _obj.shift().values).cumsum()).max().values

    if isinstance(_obj, pd.Series):
        return pd.Series(index=index, data=data, dtype=_obj.dtype, name=_obj.name)
    elif isinstance(_obj, pd.DataFrame):
        return pd.DataFrame(index=index, data=data, dtype=_obj.dtype)


def pivot(_obj, index=None, aggfunc="max", *args, **kwargs):
    """
    Reshape time-series data.

    Parameters
    ----------
    self: a Pandas Series object with DateTime index.
    rule: reshaping rule, default 'week'
        Indicates the new index level, 'year', 'week', 'day', 'hour',
        'minute' or 'second'. The lower levels are converted to columns.
    aggfunc: the Pandas.pivot_table aggfunc, default 'max'
    
    Returns
    -------
    a Pandas DataFrame object

    """
    levels = ["year", "week", "day", "hour", "minute", "second"]

    # get the gcd in seconds
    gcd = np.gcd.reduce(
        [
            int(x / np.timedelta64(1, "s"))
            for x in (_obj.index[1:] - _obj.index[:-1]).values
        ]
    )

    if gcd < 60:
        level = "second"
    elif gcd < 60 * 60:
        level = "minute"
    elif gcd < 60 * 60 * 24:
        level = "hour"
    elif gcd < 60 * 60 * 24 * 7:
        level = "day"
    else:
        level = "week"

    frame = pd.DataFrame(index=_obj.index, data={"values": _obj.values})
    frame.loc[:, ["year", "week", "day"]] = frame.index.isocalendar().values
    frame["hour"] = frame.index.hour
    frame["minute"] = frame.index.minute
    frame["second"] = frame.index.second
    result = frame.pivot_table(
        index=levels[: levels.index(index) + 1],
        columns=levels[levels.index(index) + 1 : levels.index(level) + 1],
        values="values",
        aggfunc=aggfunc,
    )

    return result


class Reshape:
    def __init__(self, _obj, rule="w", level="m", aggfunc="max", **kwargs):
        self._obj = _obj
        self.rule = _get_rule(rule)
        self.level = _get_rule(level)
        self.levels = ["year", "week", "day", "hour", "minute", "second"]
        self.aggfunc = aggfunc
        self.kwargs = kwargs
        self.calculate()

    def calculate(self, **kwargs):

        if self.levels.index(self.rule) > self.levels.index(self.level):
            raise ValueError(
                "The 'rule' argument must be greater than or equal to the 'level' argument."
            )

        frame = pd.DataFrame(self._obj)
        frame.loc[:, ["year", "week", "day"]] = frame.index.isocalendar().values
        frame["hour"] = frame.index.hour
        frame["minute"] = frame.index.minute
        frame["second"] = frame.index.second
        self.result = frame.pivot_table(
            index=self.levels[: self.levels.index(self.rule) + 1],
            columns=self.levels[
                self.levels.index(self.rule) + 1 : self.levels.index(self.level) + 1
            ],
            values=self._obj.name,
            aggfunc=self.aggfunc,
        )

    def nsmallest(self, n=1):
        """
            Reshape the Timeline object and return a pandas Series for one period.
            The period is constructed as the percentile for every 'level' time slot
            over one 'rule' period.
        
            Parameters
            ----------
            self: Bluebelt Timeline object
            n: the n-th smallest value for every 'level' timeframe, default 1
            rule: reshaping rule, default 'week'
                Indicates the new index level, 'year', 'week', 'day', 'hour',
                'minute' or 'second'. The lower levels are converted to columns.
            level: the lowest level to return, default 'minute'
            
            Returns
            -------
            a Pandas DataFrame object

        """

        # yes, reshape to the lowest level and return a series
        result = Reshape(self._obj, rule=self.level).result.iloc[:, 0]

        # set groupby_levels
        groupby_levels = self.levels[
            self.levels.index(self.rule) + 1 : self.levels.index(self.level) + 1
        ]

        # and now groupby levels down to 'rule'
        result = (
            result.groupby(level=groupby_levels, group_keys=False)
            .nsmallest(n)
            .groupby(level=groupby_levels, group_keys=False)
            .nlargest(1)
        )

        return result

    def nlargest(self, n=1):
        """
            Reshape the Timeline object and return a pandas Series for one period.
            The period is constructed as the percentile for every 'level' time slot
            over one 'rule' period.
        
            Parameters
            ----------
            self: Bluebelt Timeline object
            n: the n-th largest value for every 'level' timeframe, default 1
            rule: reshaping rule, default 'week'
                Indicates the new index level, 'year', 'week', 'day', 'hour',
                'minute' or 'second'. The lower levels are converted to columns.
            level: the lowest level to return, default 'minute'
            
            Returns
            -------
            a Pandas DataFrame object

        """
        # yes, reshape to the lowest level and return a series
        result = Reshape(self._obj, rule=self.level).result.iloc[:, 0]

        # set groupby_levels
        groupby_levels = self.levels[
            self.levels.index(self.rule) + 1 : self.levels.index(self.level) + 1
        ]

        # and now groupby levels down to 'rule'
        result = (
            result.groupby(level=groupby_levels, group_keys=False)
            .nlargest(n)
            .groupby(level=groupby_levels, group_keys=False)
            .nsmallest(1)
        )

        return result

    def __repr__(self):
        return self.result.__repr__()


def _get_rule(rule):
    if isinstance(rule, str):
        rule = rule.replace(" ", "").lower()
        unit = rule.lstrip("0123456789")
        count = int(rule[0 : len(rule) - len(unit)]) if len(rule) > len(unit) else 1
        if unit in {"s", "sec", "second", "seconds"}:
            return "second"
        elif unit in {"m", "min", "minute", "minutes"}:
            return "minute"
        elif unit in {"h", "hr", "hrs", "hour", "hours"}:
            return "hour"
        elif unit in {"d", "day", "days"}:
            return "day"
        elif unit in {"w", "wk", "week", "weeks"}:
            return "week"
        elif unit in {"y", "yr", "year", "years"}:
            return "year"
        else:
            raise ValueError(
                "Value must be year-, week-, day-, hour-, minute- or second-like."
            )
    else:
        return rule



import pandas as pd
import numpy as np

import datetime
import itertools
import math

import bluebelt.helpers.check as check
import bluebelt.helpers.convert

NoneType = type(None)


def from_dataframe(_obj, start=None, end=None, value=None, **kwargs):
    """
    Create a pandas Series timeline from a pandas DataFrame

    Parameters
    ----------
    _obj: the pandas DataFrame
    start: the column name with the start Timestamp data, default None
    end: the column name with the end Timestamp data, default None
    value: the columns with the value data, default None

    Returns
    -------
    The new pandas Series object

    """
    check.is_frame(_obj)
    check.is_str(start, end)
    check.is_str_or_none(value)

    # remove rows with start == end
    # remove NaT values
    _obj = _obj[_obj[start] != _obj[end]]
    _obj = _obj[~(_obj[start].isna() | _obj[end].isna())]

    # set new index
    index = pd.Index(_obj[start].values).union(pd.Index(_obj[end].values))

    # build input series
    if value:
        series = _obj.groupby([start, end])[value].sum().rename("value")
    else:
        series = _obj.groupby([start, end])[start].count().rename("value")

    # remove start == end
    series = series[
        series.index.get_level_values(0) != series.index.get_level_values(1)
    ]

    # remove NaT
    series = series[~series.isna()]

    # start with an empty array
    array = np.zeros(index.size)

    # add each row of the series to the array
    for key, value in series.to_dict().items():
        array = np.add(
            array,
            np.multiply(
                ((index.values >= key[0]) & (index.values < key[1])).astype(float),
                value,
            ),
        )

    result = pd.Series(index=index, data=array)

    # remove duplicate indices
    result = result[~result.index.duplicated(keep="first")]

    return result


def add_series(_obj, other=None, **kwargs):
    """
    Add data to pandas Series to build a timeline.

    Parameters
    ----------
    _obj: the pandas Series object
    other: the pandas Series to add, default None

    Returns
    -------
    The new pandas Series object

    """
    check.is_series(other)

    # build new index
    index = _obj.index.union(other.index)

    # reindex to new index
    t1 = _obj.reindex(index).ffill().fillna(0)
    t2 = other.reindex(index).ffill().fillna(0)

    return t1.add(t2)


def add_dataframe(_obj, other=None, start=None, end=None, value=None, **kwargs):
    """
    Add data to pandas Series to build a timeline.

    Parameters
    ----------
    _obj: the pandas Series
    other: the pandas DataFrame to add
    start: the column name with the start Timestamp data, default None
    end: the column name with the end Timestamp data, default None
    value: the columns with the value data, default None

    Returns
    -------
    The new pandas Series object

    """
    check.is_frame(other)
    check.is_str(start, end, value)

    other = from_dataframe(other, start=start, end=end, value=value, **kwargs)
    return _obj.reindex(other.index).ffill().fillna(0).add(other)


add_frame = add_dataframe


def add_list(_obj, other, **kwargs):
    """
    Add data to pandas Series to build a timeline.

    Parameters
    ----------
    _obj: the pandas series
    other: the object to add, default None
        The object must be a list, a tuple or a numpy ndarray.
        The first two items in the list must be or be convertible to
        Pandas Timestamps. If the list has a third item the value will
        be the third items value, if not it will be 1.
        
    Returns
    -------
    The new pandas Series object

    """
    check.is_list_or_tuple_or_array(other)

    # check if there are at least two timestamps or force an error
    start = pd.Timestamp(other[0])
    end = pd.Timestamp(other[1])

    # set value
    value = other[2] if len(other) > 2 else 1

    other_index = pd.Index([start, end], dtype="datetime64[ns]")
    new_index = _obj.index.union(other_index)

    # reindex the series
    t1 = _obj.reindex(new_index).ffill().fillna(0)
    t2 = (
        pd.Series(index=other_index, data=[value, 0],)
        .reindex(new_index)
        .ffill()
        .fillna(0)
    )

    return t1.add(t2)


add_tuple = add_list
add_array = add_list


def add_dict(_obj, other, **kwargs):
    """
    Add data to pandas Series to build a timeline.

    Parameters
    ----------
    _obj: the pandas series
    other: the dict to add, default None
        The dict must have the keys 'start' and 'end' that have Pandas
        Timestamp-like values If the 'value' key is not in the dict the
        value will be 1.
        
    Returns
    -------
    The new pandas Series object
    """

    check.is_dict(other)

    # check if there are at least two timestamps or force an error
    start = pd.Timestamp(other.get("start"))
    end = pd.Timestamp(other.get("end"))
    value = other.get("value", 1)

    other_index = pd.Index([start, end], dtype="datetime64[ns]")
    new_index = _obj.index.union(other_index)

    # reindex the Timeline objects (returns pd.Series objects)
    t1 = _obj.reindex(new_index).ffill().fillna(0)
    t2 = (
        pd.Series(index=other_index, data=[value, 0],)
        .reindex(new_index)
        .ffill()
        .fillna(0)
    )

    return t1.add(t2)


def add(_obj, other=None, start=None, end=None, value=None, **kwargs):
    """
    Add data to pandas Series to build a timeline.

    Parameters
    ----------
    _obj: the pandas series
    other: the object to add, default None
        If the object is a pandas Series a new Series is created with
        a combined index.
        If the object is a Pandas Dataframe and 'start' and 'end' are
        provided every row in the dataframe is added to the timeline. A
        'value' column is optional.
        If the object is a tuple or list and the first two items in it are
        Pandas Timestamps or can be converted to Pandas Timestamps a new
        object is created with a combined index. If a third item is given
        the value will be the third items value, if not it will be 1.
        If the object is a dict with keys 'start' and 'end' that have
        Pandas Timestamp-like values a new object is created with a combined
        index. If a 'value' key is not in the dict the value will be 1.
        If the object is a Pandas Series of int the default Pandas add
        function is used.
    start: the column name with the start Timestamp data, default None
        Only used if 'other' is a Pandas DataFrame
    end: the column name with the end Timestamp data, default None
        Only used if 'other' is a Pandas DataFrame
    value: the columns with the value data, default None
        Only used if 'other' is a Pandas DataFrame

    Returns
    -------
    The new pandas Series object

    """
    if isinstance(other, pd.Series):
        return add_series(_obj, other)

    elif isinstance(other, (list, tuple, dict, np.ndarray)):
        return add_array(_obj, other)

    elif (
        isinstance(other, pd.DataFrame, start=start, end=end, value=value)
        and start is not None
        and end is not None
    ):
        return add_dataframe(_obj, other)

    else:
        return _obj


def get_shifts(
    _obj,
    shifts=None,
    year=None,
    week=None,
    day=None,
    adjust_midnight=0,
    copy=False,
    freq=None,
    *args,
    **kwargs,
):
    """
    Get the ideal combination of shifts for a given timeline.

    Parameters
    ----------
    _obj: the pandas series
    shifts: a list of shift lengths in hours, default None
        A list of possible shift lengths can be passed in any order.
        e.g. [8, 6, 3, 4]
        Another option is to pass a list of pandas Timedelta values.
        e.g. [pandas.Timedelta('8h'), pandas.Timedelta('4.5h')]
    year: the iso year number as a filter for the pandas series, default None
        If week is not given the entire pandas series will be calculated.
        This might take some time. 
    week: the iso week number as a filter for the pandas series, default None
        If week is not given the entire pandas series will be calculated.
        This might take some time.
    freq: the frequency to use in the timeline
    adjust_midnight: adjust midnight in hours, default 0
        Every next iso day will start at midnight + adjust_midnight hours.
        e.g. adjust_midnight = -1 => the next iso day starts at 23:00

    Returns
    -------
    A new pandas DataFrame with all the shifts.

    """

    if isinstance(_obj.index, pd.MultiIndex) and all(
        [x in _obj.index.names for x in ["day", "hour", "minute"]]
    ):
        _obj.index = _obj.index.map(
            lambda x: datetime.datetime.combine(
                datetime.date.fromisocalendar(1973, 16, x[0]), datetime.time(x[1], x[2])
            )
        )
        _obj.index.freq = min(_obj.index[1:] - _obj.index[:-1])
    elif not isinstance(_obj.index, pd.DatetimeIndex):
        raise ValueError(
            f"Index must be a DatetimeIndex or MultiIndex with 'day', 'hour' and 'minute' not {type(_obj.index).__name__}."
        )

    freq = freq or pd.Timedelta(_obj.index.freq.freqstr)
    adjust = int(pd.Timedelta(f"{-adjust_midnight} hours") / freq)

    result = {}

    # filter _obj
    if year and isinstance(year, str):
        year = bluebelt.helpers.convert.str_to_list_of_int(year)
    if year and isinstance(year, int):
        _obj = _obj[_obj.index.isocalendar().year == year]
    elif year and isinstance(year, (list, tuple)):
        _obj = _obj[_obj.index.isocalendar().year.isin(year)]

    if week and isinstance(week, str):
        week = bluebelt.helpers.convert.str_to_list_of_int(week)
    if week and isinstance(week, int):
        _obj = _obj[_obj.index.isocalendar().week == week]
    elif week and isinstance(week, (list, tuple)):
        _obj = _obj[_obj.index.isocalendar().week.isin(week)]

    if day and isinstance(day, str):
        day = bluebelt.helpers.convert.str_to_list_of_int(day)
    if day and isinstance(day, int):
        _obj = _obj[_obj.index.isocalendar().day == day]
    elif day and isinstance(day, (list, tuple)):
        _obj = _obj[_obj.index.isocalendar().day.isin(day)]

    # make sure shifts is a list of pd.Timedelta
    if all([isinstance(x, (int, float)) for x in shifts]):
        shifts = [pd.Timedelta(f"{x} hours") for x in shifts]

    # loop years
    for year in _obj.index.isocalendar().year.unique():
        result[year] = {}

        # loop weeks
        for week in _obj.index.isocalendar().week.unique():
            result[year][week] = {}

            # loop weekdays
            for day in _obj.index.isocalendar().day.unique():
                arr = _obj.shift(adjust)[
                    (_obj.index.isocalendar().year == year)
                    & (_obj.index.isocalendar().week == week)
                    & (_obj.index.isocalendar().day == day)
                ]

                # if there is a day
                day_length = arr[arr.shift() > 0].index.max() - arr[arr > 0].index.min()
                result[year][week][day] = {}

                if not pd.isna(day_length):

                    # build combinations for current day
                    shift_combinations = []
                    for r in range(
                        1, math.ceil(day_length / min(shifts)) + 1
                    ):  # just a bit more than one day_size
                        shift_combinations += [
                            x
                            for x in itertools.combinations_with_replacement(shifts, r)
                            if sum(sorted(x)[1:], datetime.timedelta(0, 0))
                            <= day_length
                        ]
                    shift_combinations = sorted(
                        shift_combinations,
                        key=lambda x: sum(x, datetime.timedelta(0, 0)),
                    )  # sort by sum of combination

                    combinations = []
                    [
                        combinations.append(shift_set)
                        for shift_set in shift_combinations
                        if sum(shift_set, datetime.timedelta(0, 0))
                        not in [
                            sum(shc, datetime.timedelta(0, 0)) for shc in combinations
                        ]
                    ]

                    # build the result layer by layer
                    for layer in range(int(arr.max())):

                        result[year][week][day][layer] = {}

                        layer_sets = np.array(
                            [
                                [a.index[0], a.index[-1],]  # start  # end
                                for a in np.split(
                                    arr, np.where(np.roll(arr, 1) == 0)[0]
                                )
                                if len(a) > 1
                            ]
                        )

                        layer_sets = _get_layer(arr, layer_sets, combinations)

                        layer_set_id = 0
                        for layer_set in layer_sets:
                            result[year][week][day][layer][layer_set_id] = {}

                            combo_starts = np.array(
                                [i for i in combinations[layer_set[3]]]
                            ).cumsum() - np.array(
                                [i for i in combinations[layer_set[3]]]
                            )
                            combo_ends = np.array(
                                [i for i in combinations[layer_set[3]]]
                            ).cumsum()

                            starts = arr.index[
                                [
                                    arr.index.get_loc(
                                        max(
                                            arr.index.min(),
                                            layer_set[0]
                                            + c
                                            - (
                                                (
                                                    (
                                                        (layer_set[4] - layer_set[2])
                                                        / freq
                                                    )
                                                    // 2
                                                )
                                                * freq
                                            ),
                                        )
                                    )
                                    for c in combo_starts
                                ]
                            ]
                            ends = arr.index[
                                [
                                    arr.index.get_loc(
                                        min(
                                            arr.index.max(),
                                            layer_set[0]
                                            + c
                                            - (
                                                (
                                                    (
                                                        (layer_set[4] - layer_set[2])
                                                        / freq
                                                    )
                                                    // 2
                                                )
                                                * freq
                                            ),
                                        )
                                    )
                                    for c in combo_ends
                                ]
                            ]

                            for i, s, e in zip(range(len(starts)), starts, ends):
                                result[year][week][day][layer][layer_set_id][i] = {}
                                result[year][week][day][layer][layer_set_id][i][
                                    "start"
                                ] = s - (adjust * freq)
                                result[year][week][day][layer][layer_set_id][i][
                                    "end"
                                ] = e - (adjust * freq)

                            layer_set_id += 1

                        # remove layer from arr
                        arr = arr.subtract(
                            pd.Series(index=arr.index, data=np.where(arr > 0, 1, 0)),
                            fill_value=0,
                        )

    frame = pd.DataFrame.from_dict(
        {
            (year, week, day, layer, layer_set_id, shift): result[year][week][day][
                layer
            ][layer_set_id][shift]
            for year in result.keys()
            for week in result[year].keys()
            for day in result[year][week].keys()
            for layer in result[year][week][day].keys()
            for layer_set_id in result[year][week][day][layer].keys()
            for shift in result[year][week][day][layer][layer_set_id].keys()
        },
        orient="index",
    )

    frame = frame.reset_index()
    frame.columns = ["year", "week", "day", "layer", "set", "shift", "start", "end"]

    return frame


def _combination_id(shift_length, combinations):
    return np.argmax(
        np.array([sum(shc, datetime.timedelta(0, 0)) for shc in combinations])
        >= shift_length
    )


def _get_layer(arr, layer, combinations):
    layer_duration = layer[:, 1] - layer[:, 0]
    layer = np.append(
        layer, layer_duration.reshape(layer.shape[0], 1), axis=1
    )  # add layer duration

    shift_combination_index = np.array(
        [_combination_id(x, combinations) for x in layer[:, 2]]
    )
    layer = np.append(
        layer, shift_combination_index.reshape(layer.shape[0], 1), axis=1
    )  # add shift_combination_index

    shift_duration = np.array(
        [
            sum(
                combinations[_combination_id(x, combinations)], datetime.timedelta(0, 0)
            )
            for x in layer[:, 2]
        ]
    )
    layer = np.append(
        layer, shift_duration.reshape(layer.shape[0], 1), axis=1
    )  # add shift duration

    layer = np.append(
        layer,
        layer[:, 0].reshape(layer.shape[0], 1)
        + shift_duration.reshape(layer.shape[0], 1),
        axis=1,
    )  # add shift start + shift duration

    layer = np.append(
        layer,
        layer[:, 0].reshape(layer.shape[0], 1)
        + np.array(
            [
                sum(
                    combinations[_combination_id(x, combinations)],
                    datetime.timedelta(0, 0),
                )
                for x in layer[:, 2]
            ]
        ).reshape(layer.shape[0], 1)
        + combinations[0],
        axis=1,
    )  # add shift start + shift duration + shift_combinations[0] duration

    if layer.shape[0] > 1:
        # a<b: start of next layer set < start of this layer + combination
        # c<d: end of next layer set < start of this layer + combination + smallest combination
        combine = np.array(
            [
                ([a < b, c < d])
                for a, b, c, d in zip(
                    layer[1:, 0], layer[:-1, 5], layer[1:, 1], layer[:-1, 6]
                )
            ]
        ).any(axis=1)

        # combine sets if applicable and get_layer
        combined = np.array(
            [(a[0], b[1]) for a, b, c in zip(layer[:-1, :], layer[1:, :], combine) if c]
        )
        if combined.shape[0] > 0:
            combined_sets = _get_layer(arr, combined, combinations)
            remaining_sets = np.delete(
                layer,
                [
                    (a, b)
                    for a, b, c in zip(
                        np.arange(0, combine.size + 1)[:-1],
                        np.arange(0, combine.size + 1)[1:],
                        combine,
                    )
                    if c
                ],
                axis=0,
            )

            # build final layer
            layer = np.append(combined_sets, remaining_sets, axis=0)
    return layer



import numpy as np
import pandas as pd
import datetime
import math

import bluebelt.data.shape

import bluebelt.helpers.check as check

import warnings

NoneType = type(None)


def resolution_methods(cls):
    def sum(self):
        result = self.grouped.sum(min_count=1)
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def mean(self):
        result = self.grouped.mean()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def var(self):
        result = self.grouped.var()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def std(self):
        result = self.grouped.std()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def min(self):
        result = self.grouped.min()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def max(self):
        result = self.grouped.max()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def count(self):
        result = self.grouped.count()
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def value_range(self):
        result = self.grouped.apply(lambda x: x.max() - x.min())
        if isinstance(self, Flatten):
            result = result.reindex_like(self._obj, method="ffill")
        return result

    def nthsmallest(self, n=0, index="week", aggfunc="max"):

        pivot = bluebelt.data.shape.pivot(
            self.grouped.apply(aggfunc), index=index, aggfunc="max"
        )

        index = pivot.columns

        array = pivot.to_numpy()
        array = np.partition(array, n, axis=0)
        array = array[n : n + 1, :].flatten()
        return pd.Series(index=index, data=array)

    def nthlargest(self, n=0, aggfunc="max"):

        # not the smallest but the opposite
        n = array.shape[0] - (n + 1)

        pivot = bluebelt.data.shape.pivot(
            self.grouped.apply(aggfunc), index=index, aggfunc="max"
        )

        index = pivot.columns

        array = pivot.to_numpy()
        array = np.partition(array, n, axis=0)
        array = array[n : n + 1, :].flatten()
        return pd.Series(index=index, data=array)

    def subseries(self, count, size=1):
        """
        Count the number of times a list of 'count' items with size 'size' fit
        in each group.

        Parameters
        ----------
        self : GroupBy or Resampling object
        count: int
            The length of the sub-series.
        size: int
            The size of each element in the sub-series.

        Returns
        -------
        Series

        Example
        -------

        GroupBy or Resampling object: pd.Series([10, 8, 3, 3, 5])
        count: 3
        size: 1

        frame.blue.resample(rule=7).subseries(3, 1)
        >>> 9

        Every step subtracts (1, 1, 1, 0, 0), (1, 1, 0, 1, 0), (1, 1, 0, 0, 1),
        (1, 0, 0, 1, 1) or (0, 0, 1, 1, 1) from the group.

        step 0: (3, 3, 5, 8, 10)
        step 1: (3, 3, 4, 7, 9)
        step 2: (3, 3, 3, 6, 8)
        step 3: (2, 3, 3, 5, 7)
        step 4: (2, 2, 3, 4, 6)
        step 5: (2, 2, 2, 3, 5)
        step 6: (1, 2, 2, 2, 4)
        step 7: (1, 1, 1, 2, 3)
        step 8: (0, 1, 1, 1, 2)
        step 9: (0, 0, 0, 1, 1)
        """

        if not isinstance(count, int):
            raise ValueError("count must be an int")
        if not isinstance(size, (float, int)):
            raise ValueError("size must be float or int")
        # if isinstance(self._obj, pd.DataFrame):
        #     data = self._obj.sum(axis=1, numeric_only=True)

        def _subseries_count(series, count=3, size=1):
            # helper for resample.subseries
            series = pd.Series(series) / size
            result = series.sum() * count
            for i in range(count, 0, -1):
                result = np.minimum(
                    result,
                    math.floor(series.nsmallest(len(series) - count + i).sum() / i),
                )
            return result

        if isinstance(self._obj, pd.Series):
            result = pd.Series(
                index=self.grouped.groups.keys(),
                data=self.grouped.apply(
                    lambda x: _subseries_count(series=x, count=count, size=size)
                ).values,
                name=f"subseries {str(count)} of {str(size)}",
            )
        elif isinstance(self._obj, pd.DataFrame):
            result = pd.DataFrame(
                index=self.grouped.groups.keys(),
                data=self.grouped.apply(
                    lambda x: _subseries_count(series=x, count=count, size=size),
                ).values,
                columns=self._obj.columns,
            )

        return result

    def diff_quantity(self, shift=1, skip_na=True, skip_zero=True):
        # if this week > last week: this week - last week / this week
        # if this week < last week: last week - this week / last week

        if isinstance(self._obj, pd.DataFrame):
            data = self._obj.sum(
                axis=1, min_count=1
            )  # min_count=1 prevents np.nan values to become zeros
        else:
            data = self._obj

        # resample the data and transpose into a DataFrame with weekdays as columns
        data = data.resample(rule=self.rule, origin=self.origin, **self.kwargs)
        data = data.apply(
            lambda x: pd.DataFrame(
                index=x.index.isocalendar().day, data=x.values,
            ).transpose()
        )

        # drop the extra index level that was created in the transpose
        data = data.droplevel(-1)

        # get the difference, but only for days where both weeks have a proper value
        if skip_na and skip_zero:
            result = (
                data[(data.shift(shift).notna()) & (data.shift(shift) != 0)]
                - data.shift(shift)[(data.notna()) & (data != 0)]
            ).sum(axis=1).abs() / np.maximum(
                data[(data.shift(shift).notna()) & (data.shift(shift) != 0)].sum(
                    axis=1
                ),
                data.shift(shift)[(data.notna()) & (data != 0)].sum(axis=1),
            )
        elif skip_na:
            result = (
                data[data.shift(shift).notna()] - data.shift(shift)[data.notna()]
            ).sum(axis=1).abs() / np.maximum(
                data[data.shift(shift).notna()].sum(axis=1),
                data.shift(shift)[data.notna()].sum(axis=1),
            )
        elif skip_zero:
            result = (data[data.shift(shift) != 0] - data.shift(shift)[data != 0]).sum(
                axis=1
            ).abs() / np.maximum(
                data[data.shift(shift) != 0].sum(axis=1),
                data.shift(shift)[data != 0].sum(axis=1),
            )
        else:
            result = (data - data.shift(shift)).sum(axis=1).abs() / np.maximum(
                data.sum(axis=1), data.shift(shift).sum(axis=1)
            )

        return result

    def diff_distribution(self, shift=1, skip_na=True, skip_zero=True):
        # only works with resample('1w') or equivalent
        if self.rule != datetime.timedelta(days=7):
            raise ValueError(
                f"The object must be resampled to weekly data not {self.rule}"
            )

        # if a DataFrame is passed then sum the columns
        if isinstance(self._obj, pd.DataFrame):
            data = self._obj.sum(axis=1, numeric_only=True)
        else:
            data = self._obj

        # resample the data and transpose into a DataFrame with weekdays as columns
        data = data.resample(rule=self.rule, origin=self.origin, **self.kwargs)
        data = data.apply(
            lambda x: pd.DataFrame(
                index=x.index.isocalendar().day, data=x.values,
            ).transpose()
        )

        # drop the extra index level that was created in the transpose
        data = data.droplevel(-1)

        # get diff_quantity to calculate the relative difference
        if skip_na and skip_zero:
            diff = (
                (
                    data[(data.shift(shift).notna()) & (data.shift(shift) != 0)]
                    - data.shift(shift)[(data.notna()) & (data != 0)].multiply(
                        data[
                            (data.shift(shift).notna()) & (data.shift(shift) != 0)
                        ].sum(axis=1)
                        / data.shift(shift)[(data.notna()) & (data != 0)].sum(axis=1),
                        axis=0,
                    )
                )
                .abs()
                .sum(axis=1, min_count=1)
            )
        elif skip_na:
            diff = (
                (
                    data[data.shift(shift).notna()]
                    - data.shift(shift)[data.notna()].multiply(
                        data[data.shift(shift).notna()].sum(axis=1)
                        / data.shift(shift)[data.notna()].sum(axis=1),
                        axis=0,
                    )
                )
                .abs()
                .sum(axis=1, min_count=1)
            )
        elif skip_zero:
            diff = (
                (
                    data[data.shift(shift) != 0]
                    - data.shift(shift)[data != 0].multiply(
                        data[data.shift(shift) != 0].sum(axis=1)
                        / data.shift(shift)[data != 0].sum(axis=1),
                        axis=0,
                    )
                )
                .abs()
                .sum(axis=1, min_count=1)
            )
        else:
            diff = (
                (
                    data
                    - data.shift(shift).multiply(
                        data.sum(axis=1) / data.shift(shift).sum(axis=1), axis=0,
                    )
                )
                .abs()
                .sum(axis=1, min_count=1)
            )

        # build result
        result = pd.Series(
            diff / (data.sum(axis=1, min_count=1) * 2), name="distribution",
        )

        return result

    def diff_skills(self, shift=1, skip_na=True, skip_zero=True):

        """
        Calculate the difference in skills distribution from week to week in
        from a resampled DataFrame.

        Parameters
        ----------
        self : GroupBy or Resampling object

        Returns
        -------
        Series

        Example
        -------
        index = pd.date_range(start='2022-03-07 00:00:00', periods=21)
        data = {'skill 1': [20,20,10,15,10,10,10,
                            20,20,10,10,10,10,10,
                            20,20,10,10,10,10,10,],
                'skill 2': [20,10,20,5,10,10,10,
                            20,20,10,10,10,10,10,
                            20,10,20,10,10,10,10,],
                }
        frame = pd.DataFrame(index=index, data=data)

        >>> frame.blue.resample(rule=7).diff_skills()

        2022-03-07         NaN
        2022-03-14    0.027778
        2022-03-21    0.000000
        Name: skills, dtype: float64
        """

        # only works with resample('1w') or equivalent
        if self.rule != datetime.timedelta(days=7):
            raise ValueError(
                f"The object must be resampled to weekly data not {self.rule}"
            )

        # if the object is a Series return Series of zeros.
        if isinstance(self._obj, pd.Series):
            return pd.Series(
                index=self.grouped.sum().index,
                data=np.zeros(self.grouped.sum().size),
                name="skills",
            )

        # resample the data and transpose into a DataFrame with weekdays as columns
        data = self.grouped.apply(
            lambda x: pd.DataFrame(
                index=x.index.isocalendar().day, data=x.values, columns=x.columns,
            ).transpose()
        )

        # build a filter for the dataframe for weekdays with value zero or np.nan
        summed = data.groupby(level=0).apply(pd.DataFrame.sum, skipna=True)

        filt_index = pd.MultiIndex.from_product([summed.index, self._obj.columns])
        filt_data = np.repeat(summed.values, repeats=self._obj.shape[1], axis=0)

        if skip_na and skip_zero:
            filt_data = (filt_data != 0) & (np.logical_not(np.isnan(filt_data)))
        elif skip_na:
            filt_data = np.logical_not(np.isnan(filt_data))
        elif skip_zero:
            filt_data = filt_data != 0
        else:
            filt_data = np.full(filt_data.shape, True)

        filt = pd.DataFrame(index=filt_index, data=filt_data, columns=summed.columns,)

        # get diff_quantity to calculate the relative difference
        if skip_na and skip_zero:
            diff_quantity = summed[
                (summed.shift(shift).notna()) & (summed.shift(shift) != 0)
            ].sum(axis=1) / summed.shift(shift)[(summed.notna()) & (summed != 0)].sum(
                axis=1
            )
        elif skip_na:
            diff_quantity = summed[summed.shift(shift).notna()].sum(
                axis=1
            ) / summed.shift(shift)[summed.notna()].sum(axis=1)
        elif skip_zero:
            diff_quantity = summed[summed.shift(shift) != 0].sum(axis=1) / summed.shift(
                shift
            )[summed != 0].sum(axis=1)
        else:
            diff_quantity = summed.sum(axis=1) / summed.shift(shift).sum(axis=1)

        # calculate diff_skills

        # filter the data with shifted filter and the shifted data with the unshifted filter
        # remember that the shift is equal to the number of skills (self._obj.shape[1])
        # sum the data for the weekdays
        # unstack the skills => the result is grouped data with skills as columns

        diff = (
            # the data with shifted filter minus the shifted data with filter
            (
                data[filt.shift(self._obj.shape[1] * shift)]
                .sum(axis=1)
                .unstack(level=1)
                - data.shift(self._obj.shape[1] * shift)[filt]
                .sum(axis=1)
                .unstack(level=1)
                .multiply(diff_quantity, axis=0)
            )
            .abs()
            .sum(axis=1, min_count=1)
        )

        # get the difference relative to the sum of the row with the shifted filter applied
        result = pd.Series(
            diff
            / (
                data[filt.shift(self._obj.shape[1] * shift)]
                .sum(axis=1)
                .unstack(level=1)
                .sum(axis=1, min_count=1)
                * 2
            ),
            name="skills",
        )

        return result

    setattr(cls, "sum", sum)
    setattr(cls, "mean", mean)
    setattr(cls, "var", var)
    setattr(cls, "std", std)
    setattr(cls, "min", min)
    setattr(cls, "max", max)
    setattr(cls, "count", count)
    setattr(cls, "value_range", value_range)
    setattr(cls, "nthsmallest", nthsmallest)
    setattr(cls, "nthlargest", nthlargest)
    setattr(cls, "subseries", subseries)
    setattr(cls, "diff_quantity", diff_quantity)
    setattr(cls, "diff_distribution", diff_distribution)
    setattr(cls, "diff_skills", diff_skills)
    return cls


@resolution_methods
class Resample:
    """
    Resample time-series data.
    The object must have a DatetimeIndex

    Parameters
    ----------
    self: Series or DataFrame
    rule: DateOffset, Timedelta or str, default 'week'
        The offset string or object representing target conversion.
        e.g. '1w', 'week', '4d', '4 days'
    ffill: Forward fill data if applicable, default None
        If 'rule' is smaller than the greatest step in the index values of self
        the NaN values will be forward filled if True. Does raise a warning to
        point out this behaviour.

    kwargs: other pandas.Series.resample keywords

    Returns
    -------
    pandas.core.Resampler

    Methods
    -------
    sum: Calculate the sum of all the values in each group.
    mean: Calculate the mean of all the values in each group.
    var: Calculate the variance of all the values in each group.
    std: Calculate the standard deviation of all the values in each group.
    min: Calculate the minimum value of all the values in each group.
    max: Calculate the maximum value of all the values in each group.
    count: Calculate the count of all the values in each group.
    value_range: Calculate the value range of all the values in each group.
    subsize: Calculate the count of subgroups with a size that fit in each
        group.
    diff_quantity: Calculate the difference of the sum of all values
        in adjecent groups of the resampled DataFrame.
    diff_distribution: Calculate the difference in weekday distribution
        from week to week in from a resampled DataFrame. The rule for
        resampling must be '1w' or equivalent
    diff_skills: Calculate the difference in skills distribution from week
        to week in from a resampled DataFrame. The rule for resampling must
        be '1w' or equivalent.

    """

    def __init__(self, _obj, rule, ffill=None, **kwargs):
        self._obj = _obj
        self.rule = _get_rule(rule)
        self.ffill = ffill
        self.gcd = np.gcd.reduce(
            [
                int(x / np.timedelta64(1, "ns"))
                for x in (_obj.index[1:] - _obj.index[:-1]).values
            ]
        )
        self.origin = _get_origin(_obj, self.rule)
        self.kwargs = kwargs
        self.calculate()

    def calculate(self, **kwargs):

        if isinstance(self.ffill, NoneType):
            # consider timeline to have closed attribute
            warnings.warn(
                f"\n\n"
                + f"The resample rule ({pd.Timedelta(self.rule)}) is smaller than the "
                + f"greatest step in the series index values ({(self._obj.index[1:] - self._obj.index[:-1]).max()}). "
                + f"This results in NaN values in the resampled series."
                + f"\nTo prevent this the gaps in the data are forward filled. To prevent this behaviour set 'ffill=False'.\n "
                + f"\nSilence this warning by setting 'ffill=True'.\n ",
                Warning,
            )

            self.ffill = True

        # check if the rule < greatest diff
        if (
            pd.Timedelta(self.rule) < (self._obj.index[1:] - self._obj.index[:-1]).max()
            and self.ffill
        ):

            self.grouped = (
                self._obj.resample(f"{self.gcd}ns")
                .ffill()
                .resample(rule=self.rule, origin=self.origin, **self.kwargs)
            )

        else:
            self.grouped = self._obj.resample(
                rule=self.rule, origin=self.origin, **self.kwargs
            )

    def __repr__(self):
        return self.grouped.__repr__()

    def apply(self, func, *args, **kwargs):
        return self.grouped.apply(func, *args, **kwargs)


class Flatten(Resample):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def __getattr__(self, attr):
        return super().__getattr__(attr).reindex_like(self._obj, method="ffill")

    # result = result.reindex_like(self._obj, method="ffill")


def _get_origin(_obj, rule):
    if (rule / datetime.timedelta(days=7)).is_integer():
        days = int(
            ((_obj.index.isocalendar().week[0] - 1) * 7) + _obj.index[0].weekday()
        )
        return _obj.index[0] - datetime.timedelta(days=days)
    elif (rule / datetime.timedelta(days=1)).is_integer():
        days = int(_obj.index[0].weekday())
        return _obj.index[0] - datetime.timedelta(days=days)
    else:
        return "start_day"


def _get_rule(rule):
    if isinstance(rule, str):
        rule = rule.replace(" ", "").lower()
        unit = rule.lstrip("0123456789")
        count = int(rule[0 : len(rule) - len(unit)]) if len(rule) > len(unit) else 1
        if unit in {"μs", "microsecond", "microseconds"}:
            return datetime.timedelta(microseconds=count)
        if unit in {"ms", "millisecond", "milliseconds"}:
            return datetime.timedelta(milliseconds=count)
        if unit in {"s", "sec", "second", "seconds"}:
            return datetime.timedelta(seconds=count)
        if unit in {"m", "min", "minute", "minutes"}:
            return datetime.timedelta(minutes=count)
        if unit in {"h", "hr", "hrs", "hour", "hours"}:
            return datetime.timedelta(hours=count)
        if unit in {"d", "day", "days"}:
            return datetime.timedelta(days=count)
        if unit in {"w", "wk", "week", "weeks"}:
            return datetime.timedelta(weeks=count)
    elif isinstance(rule, int):
        return datetime.timedelta(days=rule)
    else:
        return rule



import pandas as pd
import numpy as np

import bluebelt.data.resolution
import bluebelt.data.shape
import bluebelt.data.timeline

import bluebelt.helpers.check as check
import bluebelt.helpers.convert
from bluebelt.helpers.io import to_pickle

from bluebelt.helpers.decorators.docstring import docstring

import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle, Patch
import seaborn as sns

import plotly.graph_objects as go
from plotly.subplots import make_subplots
import bluebelt.helpers.matplotlib2plotly as m2p

import copy
import datetime
import os


class Timeline:
    def __init__(
        self,
        other,
        start=None,
        end=None,
        value=None,
        name=None,
        dtype=None,
        *args,
        **kwargs,
    ):

        if isinstance(other, pd.Series):
            if isinstance(other.index, pd.MultiIndex) and all(
                [x in other.index.names for x in ["day", "hour", "minute"]]
            ):
                other.index = other.index.map(
                    lambda x: datetime.datetime.combine(
                        datetime.date.fromisocalendar(1973, 16, x[0]),
                        datetime.time(x[1], x[2]),
                    )
                )
                other.index.freq = min(other.index[1:] - other.index[:-1])

            elif not isinstance(other.index, pd.DatetimeIndex):
                raise ValueError(
                    f"Index must be a DatetimeIndex or MultiIndex with 'day', 'hour' and 'minute' not {type(other.index).__name__}."
                )
            self._obj = other

        elif isinstance(other, pd.DataFrame):
            check.is_str(start, end)
            check.is_str_or_none(value)

            # clean the dataframe
            other = other[other[start] != other[end]]  # remove rows with start == end
            other = other[
                ~(other[start].isna() | other[end].isna())
            ]  # remove NaT values

            # set new index
            index = (
                pd.Index(other[start].values)
                .union(pd.Index(other[end].values))
                .drop_duplicates(keep="first")
            )

            # build input series
            if value:
                series = other.groupby([start, end])[value].sum().rename("value")
            else:
                series = other.groupby([start, end])[start].count().rename("value")

            # remove start == end
            series = series[
                series.index.get_level_values(0) != series.index.get_level_values(1)
            ]

            # remove NaT
            series = series[~series.isna()]

            # start with an empty array
            array = np.zeros(index.size)

            # add each row of the series to the array
            for key, value in series.to_dict().items():
                array = np.add(
                    array,
                    np.multiply(
                        ((index.values >= key[0]) & (index.values < key[1])).astype(
                            float
                        ),
                        value,
                    ),
                )

            result = pd.Series(
                index=index, data=array, name=name, dtype=(dtype or array.dtype)
            )

            # remove duplicate indices
            result = result[~result.index.duplicated(keep="first")]

            self._obj = result

    def __repr__(self):
        print("Timeline")
        return self._obj.__repr__()

    def __getattr__(self, attr):
        return self._obj.__getattr__(attr)

    def __add__(self, other):
        index = self._obj.index.union(other._obj.index)
        t1 = self._obj.reindex(index).ffill().fillna(0)
        t2 = other._obj.reindex(index).ffill().fillna(0)
        return Timeline(t1.add(t2))

    def __sub__(self, other):
        index = self._obj.index.union(other._obj.index)
        t1 = self._obj.reindex(index).ffill().fillna(0)
        t2 = other._obj.reindex(index).ffill().fillna(0)

        result = t1.subtract(t2)
        return Timeline(result.where(result > 0, 0))

    def to_pickle(self, path):
        to_pickle(self, path)

    @docstring(bluebelt.data.shape.reduce)
    def reduce(self, *args, **kwargs):
        return Timeline(bluebelt.data.shape.reduce(self._obj, **kwargs))

    @docstring(bluebelt.data.resolution.Resample)
    def resample(self, rule="15m", ffill=True, aggfunc="max", *args, **kwargs):
        result = bluebelt.data.resolution.Resample(self._obj, rule, ffill=ffill)
        return Timeline(getattr(result, aggfunc)(*args, **kwargs))

    @docstring(bluebelt.data.shape.pivot)
    def pivot(self, index="week", aggfunc="max", **kwargs):
        check.has_datetimeindex(self._obj)
        return bluebelt.data.shape.pivot(
            self._obj, index=index, aggfunc=aggfunc, **kwargs
        )

    @docstring(bluebelt.data.projection.project)
    def project(self, year=None, adjust_holidays=True):
        return Timeline(
            bluebelt.data.projection.project(
                self._obj, year=year, adjust_holidays=adjust_holidays
            )
        )

    def reshape(self, rule="week", level="minute", aggfunc="max"):
        """
            Reshape the Timeline object.
        
            Parameters
            ----------
            self: Bluebelt Timeline object
            rule: reshaping rule, default 'week'
                Indicates the new index level, 'year', 'week', 'day', 'hour',
                'minute' or 'second'. The lower levels are converted to columns.
            level: the lowest level to return, default 'minute'
            aggfunc: the Pandas.pivot_table aggfunc, default 'max'
            
            Returns
            -------
            a Pandas DataFrame object

        """
        name = self._obj.name
        levels = ["year", "week", "day", "hour", "minute", "second"]

        if not (rule in levels and level in levels):
            raise ValueError(
                "Both rule and level should be 'year', 'week', 'day', 'hour', 'minute' or 'second'"
            )

        # avoid trouble if the level has a lower index than rule
        level = level if levels.index(rule) < levels.index(level) else rule

        frame = pd.DataFrame(self._obj)
        frame.loc[:, ["year", "week", "day"]] = frame.index.isocalendar()
        frame["hour"] = frame.index.hour
        frame["minute"] = frame.index.minute
        frame["second"] = frame.index.second
        return frame.pivot_table(
            index=levels[: levels.index(rule) + 1],
            columns=levels[levels.index(rule) + 1 : levels.index(level) + 1],
            values=name,
            aggfunc=aggfunc,
        )

    def percentile(self, percentile=0.1, how="smallest", rule="week", level="minute"):
        """
            Reshape the Timeline object and return a pandas Series for one period.
            The period is constructed as the percentile for every 'level' time slot
            over one 'rule' period.
        
            Parameters
            ----------
            self: Bluebelt Timeline object
            percentile: the percentile to calculate, default 0.1
            how: calculate the percentile as 'smallest' or 'largest', default 'smallest'
            rule: reshaping rule, default 'week'
                Indicates the new index level, 'year', 'week', 'day', 'hour',
                'minute' or 'second'. The lower levels are converted to columns.
            level: the lowest level to return, default 'minute'
            
            Returns
            -------
            a Pandas DataFrame object

        """
        levels = ["year", "week", "day", "hour", "minute", "second"]

        # avoid trouble if the level has a lower index than rule
        level = (
            level
            if levels.index(rule) < levels.index(level)
            else levels[levels.index(rule) + 1]
        )

        # pivot to the lowest level and return a series
        result = self.pivot(index=level).max(axis=1)

        # get n from percentile
        n = np.round(
            result.groupby(level=levels[: levels.index(rule) + 1]).ngroups * percentile
        ).astype(int)

        # set groupby_levels
        groupby_levels = levels[levels.index(rule) + 1 : levels.index(level) + 1]

        # and now groupby levels down to 'rule'
        if how == "largest":
            result = (
                result.groupby(level=groupby_levels, group_keys=False)
                .nlargest(n)
                .groupby(level=groupby_levels, group_keys=False)
                .nsmallest(1)
            )
        elif how == "smallest":
            result = (
                result.groupby(level=groupby_levels, group_keys=False)
                .nsmallest(n)
                .groupby(level=groupby_levels, group_keys=False)
                .nlargest(1)
            )
        else:
            raise ValueError(
                f"parameter 'how' must be 'smallest' or 'largest', not {how}."
            )

        return result

    def nthsmallest(self, n=1, rule="week", level="minute"):
        """
            Reshape the Timeline object and return a pandas Series for one period.
            The period is constructed as the percentile for every 'level' time slot
            over one 'rule' period.
        
            Parameters
            ----------
            self: Bluebelt Timeline object
            n: the n-th smallest value for every 'level' timeframe, default 1
            rule: reshaping rule, default 'week'
                Indicates the new index level, 'year', 'week', 'day', 'hour',
                'minute' or 'second'. The lower levels are converted to columns.
            level: the lowest level to return, default 'minute'
            
            Returns
            -------
            a Pandas DataFrame object

        """
        levels = ["year", "week", "day", "hour", "minute", "second"]

        # avoid trouble if the level has a lower index than rule
        level = (
            level
            if levels.index(rule) < levels.index(level)
            else levels[levels.index(rule) + 1]
        )

        # pivot to the lowest level and return a series
        result = self.pivot(index=level).max(axis=1)

        # set groupby_levels
        groupby_levels = levels[levels.index(rule) + 1 : levels.index(level) + 1]

        # and now groupby levels down to 'rule'
        result = (
            result.groupby(level=groupby_levels, group_keys=False)
            .nsmallest(n)
            .groupby(level=groupby_levels, group_keys=False)
            .max()
        )

        return result

    nsmallest = nthsmallest

    def nthlargest(self, n=1, rule="week", level="minute"):
        """
            Reshape the Timeline object and return a pandas Series for one period.
            The period is constructed as the percentile for every 'level' time slot
            over one 'rule' period.
        
            Parameters
            ----------
            self: Bluebelt Timeline object
            n: the n-th largest value for every 'level' timeframe, default 1
            rule: reshaping rule, default 'week'
                Indicates the new index level, 'year', 'week', 'day', 'hour',
                'minute' or 'second'. The lower levels are converted to columns.
            level: the lowest level to return, default 'minute'
            
            Returns
            -------
            a Pandas DataFrame object

        """
        levels = ["year", "week", "day", "hour", "minute", "second"]

        # avoid trouble if the level has a lower index than rule
        level = (
            level
            if levels.index(rule) < levels.index(level)
            else levels[levels.index(rule) + 1]
        )

        # pivot to the lowest level and return a series
        result = self.pivot(index=level).max(axis=1)

        # set groupby_levels
        groupby_levels = levels[levels.index(rule) + 1 : levels.index(level) + 1]

        # and now groupby levels down to 'rule'
        result = (
            result.groupby(level=groupby_levels, group_keys=False)
            .nlargest(n)
            .groupby(level=groupby_levels, group_keys=False)
            .min()
        )

        return result

    nlargest = nthlargest

    def filter(self, year=None, week=None, day=None, *args, **kwargs):

        _obj = copy.copy(self._obj)

        if year and isinstance(year, str):
            year = bluebelt.helpers.convert.str_to_list_of_int(year)
        if year and isinstance(year, int):
            _obj = _obj[_obj.index.isocalendar().year == year]
        elif year and isinstance(year, (list, tuple)):
            _obj = _obj[_obj.index.isocalendar().year.isin(year)]

        if week and isinstance(week, str):
            week = bluebelt.helpers.convert.str_to_list_of_int(week)
        if week and isinstance(week, int):
            _obj = _obj[_obj.index.isocalendar().week == week]
        elif week and isinstance(week, (list, tuple)):
            _obj = _obj[_obj.index.isocalendar().week.isin(week)]

        if day and isinstance(day, str):
            day = bluebelt.helpers.convert.str_to_list_of_int(day)
        if day and isinstance(day, int):
            _obj = _obj[_obj.index.isocalendar().day == day]
        elif day and isinstance(day, (list, tuple)):
            _obj = _obj[_obj.index.isocalendar().day.isin(day)]

        return Timeline(_obj)

    def standard_week(
        self,
        how="nthsmallest",
        n=1,
        level="minute",
        year=None,
        week=None,
        day=None,
        *args,
        **kwargs,
    ):

        if how in ["nsmallest", "nthsmallest"]:
            _sw = self.nthsmallest(n=n, rule="week", level=level).to_dict()
        elif how in ["nlargest", "nthlargest"]:
            _sw = self.nthlargest(n=n, rule="week", level=level).to_dict()
        else:
            raise ValueError(
                f"'how' must be 'nthsmallest' or 'nthlargest', not '{how}'."
            )

        return Timeline(
            self.filter(year=year, week=week, day=day, *args, **kwargs)
            ._obj.index.to_series()
            .apply(lambda x: _sw.get((x.isocalendar().weekday, x.hour, x.minute)))
        )

    @docstring(bluebelt.data.timeline.get_shifts)
    def get_shifts(
        self,
        shifts=None,
        year=None,
        week=None,
        day=None,
        adjust_midnight=0,
        copy=False,
        freq=None,
        *args,
        **kwargs,
    ):
        return TimelineShifts(
            self._obj,
            shifts=shifts,
            year=year,
            week=week,
            day=day,
            adjust_midnight=adjust_midnight,
            copy=copy,
            freq=freq,
            *args,
            **kwargs,
        )


class TimelineShifts:
    def __init__(
        self,
        _obj,
        shifts=None,
        year=None,
        week=None,
        day=None,
        adjust_midnight=0,
        copy=False,
        freq=None,
        *args,
        **kwargs,
    ):
        self.shifts = shifts
        self.year = year
        self.week = week
        self.day = day
        self.adjust_midnight = adjust_midnight
        self.copy = copy

        # filter _obj
        if freq is None and _obj.index.freq is None:
            freq = self._get_freq()
        elif freq is None:
            freq = freq or pd.Timedelta(_obj.index.freq.freqstr)

        if year and isinstance(year, str):
            year = bluebelt.helpers.convert.str_to_list_of_int(year)
        if year and isinstance(year, int):
            _obj = _obj[_obj.index.isocalendar().year == year]
        elif year and isinstance(year, (list, tuple)):
            _obj = _obj[_obj.index.isocalendar().year.isin(year)]

        if week and isinstance(week, str):
            week = bluebelt.helpers.convert.str_to_list_of_int(week)
        if week and isinstance(week, int):
            _obj = _obj[_obj.index.isocalendar().week == week]
        elif week and isinstance(week, (list, tuple)):
            _obj = _obj[_obj.index.isocalendar().week.isin(week)]

        if day and isinstance(day, str):
            day = bluebelt.helpers.convert.str_to_list_of_int(day)
        if day and isinstance(day, int):
            _obj = _obj[_obj.index.isocalendar().day == day]
        elif day and isinstance(day, (list, tuple)):
            _obj = _obj[_obj.index.isocalendar().day.isin(day)]

        self._obj = _obj
        self.frame = bluebelt.data.timeline.get_shifts(
            self._obj,
            shifts=shifts,
            year=year,
            week=week,
            adjust_midnight=adjust_midnight,
            copy=copy,
            freq=freq,
            *args,
            **kwargs,
        )

    def __repr__(self):
        return self.frame.__repr__()

    def _repr_html_(self):
        return self.frame._repr_html_()

    def _get_freq(self):

        return np.gcd.reduce(
            [
                int(x / np.timedelta64(1, "ns"))
                for x in (self._obj.index[1:] - self._obj.index[:-1]).values
            ]
        )

    def to_pickle(self, path):
        to_pickle(self, path)

    def plot(
        self,
        xlim=(None, None),
        ylim=(None, None),
        max_xticks=None,
        format_xticks=None,
        format_yticks=None,
        title=None,
        xlabel=None,
        ylabel=None,
        palette=None,
        legend=True,
        path=None,
        **kwargs,
    ):

        palette = palette or bluebelt.style("colors")[1:]

        if bluebelt.config("plotting") == "matplotlib":

            colors = {
                s: c for s, c in zip(self.shifts, sns.color_palette(palette).as_hex())
            }

            # prepare figure
            fig, ax = plt.subplots(nrows=1, ncols=1, **kwargs)

            ax.plot(self._obj, **bluebelt.style(f"line"))

            for i, row in self.frame.iterrows():
                ax.add_patch(
                    Rectangle(
                        (row["start"], row["layer"]),
                        row["end"] - row["start"],
                        1,
                        color=colors.get(
                            (row["end"] - row["start"]) / datetime.timedelta(minutes=60)
                        ),
                        lw=0,
                    )
                )

            # limit axis
            ax.set_xlim(xlim)
            ax.set_ylim(ylim)

            # set ticks
            bluebelt.helpers.ticks.year_week(self._obj, ax=ax, max_xticks=max_xticks)

            # format ticks
            if format_xticks:
                ax.set_xticks(ax.get_xticks())
                ax.set_xticklabels([f"{x:{format_xticks}}" for x in ax.get_xticks()])
            if format_yticks:
                ax.set_yticks(ax.get_yticks())
                ax.set_yticklabels([f"{y:{format_yticks}}" for y in ax.get_yticks()])

            # labels
            if title:
                ax.set_title(title)
            if xlabel:
                ax.set_xlabel(xlabel)
            if ylabel:
                ax.set_ylabel(ylabel)

            # legend
            if legend:
                legend_elements = [
                    Patch(facecolor=c, label=f"{s} hour shift")
                    for s, c in zip(self.shifts, sns.color_palette(palette).as_hex())
                ]
                ax.legend(handles=legend_elements, loc="best")
            elif ax.get_legend() is not None:
                ax.get_legend().set_visible(False)

            plt.tight_layout()

            # file
            if path:
                if len(os.path.dirname(path)) > 0 and not os.path.exists(
                    os.path.dirname(path)
                ):
                    os.makedirs(os.path.dirname(path))
                plt.savefig(path)
                plt.close()
            else:
                plt.close()
                return fig

        elif bluebelt.config("plotting") == "plotly":

            layout = go.Layout(
                m2p.layout(
                    title=title,
                    xlabel=xlabel,
                    ylabel=ylabel,
                    xlim=xlim,
                    ylim=ylim,
                    format_xticks=format_xticks,
                    format_yticks=format_yticks,
                    legend=legend,
                    **kwargs,
                )
            )

            data = [
                dict(
                    type="scatter",
                    mode="lines",
                    x=np.array(
                        [
                            [row["start"], row["end"], row["end"], row["start"], None]
                            for i, row in self.frame.iterrows()
                            if (row["end"] - row["start"]) / pd.Timedelta("1h") == shift
                        ]
                    ).flatten()[:-1],
                    y=np.array(
                        [
                            [
                                row["layer"],
                                row["layer"],
                                row["layer"] + 1,
                                row["layer"] + 1,
                                None,
                            ]
                            for i, row in self.frame.iterrows()
                            if (row["end"] - row["start"]) / pd.Timedelta("1h") == shift
                        ]
                    ).flatten()[:-1],
                    name=f"{shift} hour shift",
                    line=dict(width=0),
                    fill="toself",
                    fillcolor=m2p.to_rgba(
                        bluebelt.style("colors")[self.shifts.index(shift) + 1], 1
                    ),
                    opacity=1,
                )
                for shift in self.shifts
            ] + [
                dict(
                    type="scatter",
                    mode="lines",
                    x=self._obj.index,
                    y=self._obj.values,
                    name=self._obj.name or "timeline",
                    line=dict(
                        color=bluebelt.style("colors")[0],
                        width=bluebelt.style("line.linewidth"),
                        dash="dot",
                    ),
                    opacity=1,
                )
            ]

            fig = go.Figure(data=data, layout=layout)

            return fig




def str_to_list_of_int(x):
    """
    Build a list of integers from a string. Usefull for making lists of weeks, days, year, ...

    Parameters
    ----------
    x: the string
    
    Returns
    -------
    a list of integers

    Example
    -------
    str_to_list_of_int('3-7, 12-17')
    >> [3, 4, 5, 6, 7, 12, 13, 14, 15, 16, 17]

    """
    return [
        item
        for sublist in [
            list(range(int(group.split("-")[0]), int(group.split("-")[1]) + 1))
            for group in x.split(",")
        ]
        for item in sublist
    ]

